import sys
import re
import argparse

from .chain import chain
from .textobject import TextObject


print_ = print


def text_generator(data, width=None):
    m = chain(data).map(TextObject)

    for lo in m:
        if width and width > 0 and lo.width > width:
            for i in lo.wrap(width):
                yield i
        else:
            yield lo


def wrap(data, width=0, padding=0, corners=".''.", edges='-|-|'):
    if width and padding:
        width -= 2 * padding
        if width <= 0:
            raise ValueError('No remain width available, increase width or decrease padding.')

    if len(corners) == 1:
        corners = corners * 4

    if len(corners) == 2:
        corners = corners + corners[::-1]

    if len(corners) != 4:
        raise ValueError('Corners should be {1, 2, 4} characters/strings.')

    corner_tr = TextObject(corners[0], rstrip=False)
    corner_br = TextObject(corners[1], rstrip=False)
    corner_bl = TextObject(corners[2], rstrip=False)
    corner_tl = TextObject(corners[3], rstrip=False)

    if len(edges) == 1:
        edges = edges * 4

    elif len(edges) == 2:
        edges = edges * 2

    elif len(edges) == 3:
        edges = edges + edges[1:2]

    elif len(edges) != 4:
        raise ValueError('Edges should be 1 ~ 4 characters/strings')

    edge_t = TextObject(edges[0], rstrip=False)
    edge_r = TextObject(edges[1], rstrip=False)
    edge_b = TextObject(edges[2], rstrip=False)
    edge_l = TextObject(edges[3], rstrip=False)

    if not chain([edge_t, edge_b]).all(lambda e: e.width > 0):
        raise ValueError('Top and bottom edge\'s width should larger than 0.')

    if not chain([edge_r, corner_tr, corner_br]).map(lambda x: x.width).all_identical():
        raise ValueError('Corners and their neighbor edges should have same width.')

    if not chain([edge_l, corner_tl, corner_bl]).map(lambda x: x.width).all_identical():
        raise ValueError('Corners and their neighbor edges should have same width.')

    ret = list(text_generator(data, width=width))
    max_width = max(*chain(ret).map(lambda lo: lo.width), width)
    edge_t = edge_t.repeat_to(max_width + 2 * padding)
    edge_b = edge_b.repeat_to(max_width + 2 * padding)
    max_width = max(max_width, edge_t.width - 2 * padding, edge_b.width - 2 * padding)

    return [
        '{ctl}{et}{ctr}'.format(
            ctl=corner_tl,
            et=edge_t,
            ctr=corner_tr,
        )] + chain(ret).map(
            lambda lo: '{el}{p}{t}{s}{p}{er}'.format(
                el=edge_l,
                p=' ' * padding,
                t=lo.text,
                s=' ' * (max_width - lo.width),
                er=edge_r,
            )
        ).list + ["{cbl}{eb}{cbr}".format(
            cbl=corner_bl,
            eb=edge_b,
            cbr=corner_br,
        )
    ]


def print(data, width=0, padding=0, corners=".''.", edges='-|-|', **kwargs):
    for l in wrap(data, width=width, padding=padding, corners=corners, edges=edges):
        print_(l, **kwargs)


def frame_argument(v):
    ret = v

    if len(ret) == 1:
        ret = ret * 4

    if len(ret) == 2:
        ret = ret + ret[::-1]

    if len(ret) != 4:
        raise argparse.ArgumentTypeError('Corners should be {1, 2, 4} characters/strings.')

    return ret



def main():
    parser = argparse.ArgumentParser(
        prog='ascii_frame',
        description='Wrap text with an ASCII frame.',
    )

    parser.add_argument('-w', '--width',
        type=int, default=0,
        help='Set the width of ASCII frame.')

    parser.add_argument('-p', '--padding',
        type=int, default=0,
        help='Set the padding of ASCII frame.')

    parser.add_argument('-f', '--input-file',
        type=argparse.FileType('r'),
        default=sys.stdin,
        help='Read content from file.')

    parser.add_argument('-c', '--corners',
        type=str, nargs='*', default=".''.",
        help='Customize corners of frame.  4 chars in clockwise order from top right.')

    parser.add_argument('-e', '--edges',
        type=str, nargs='*', default="-|-|",
        help='Customize corners of frame.  4 chars in clockwise order from top.')

    args = parser.parse_args()
    print(args.input_file,
        width=args.width,
        padding=args.padding,
        corners=args.corners,
        edges=args.edges)



class chain:
    def __init__(self, value):
        self.value = value
        for attr in dir(value):
            if not attr.startswith('__'):
                self.__dict__[attr] = getattr(self.value, attr)

    def __iter__(self):
        for i in self.value:
            yield i

    def __repr__(self):
        return repr(self.value)

    def __str__(self):
        return str(self.value)

    def __add__(self, other):
        if isinstance(other, chain):
            return chain(self.value + other.value)
        return chain(self.value + other)

    def __sub__(self, other):
        if isinstance(other, chain):
            return chain(self.value - other.value)
        return chain(self.value - other)

    def map(self, func):
        return chain(map(func, self.value))

    def filter(self, func):
        return chain(filter(func, self.value))

    def all(self, func):
        return all(self.map(func))

    def all_identical(self):
        return len(set(self.value)) <= 1

    @property
    def list(self):
        return list(self)

    @property
    def dict(self):
        return dict(self)

    @property
    def set(self):
        return set(self)



from ..textobject import TextObject


def test_textobject_text():
    assert TextObject('normal').text == 'normal'
    assert TextObject('right  ').text == 'right'
    assert TextObject('\033[1;31mColor!\033[m   ').text == '\033[1;31mColor!\033[m'
    assert TextObject('right  ', rstrip=False).text == 'right  '


def test_textobject_width():
    assert TextObject('normal').width == 6
    assert TextObject('中文').width == 4
    assert TextObject('\033[1;31mColor!\033[m   ').width == len('Color!')
    assert TextObject('\033[1;31m中文\033[m   ').width == TextObject('中文').width


def test_textobject_eq_str():
    assert TextObject('normal') == TextObject('normal')
    assert TextObject('normal') == 'normal'


def test_textobject_wrap():
    assert list(TextObject('normal text').wrap(100)) == ['normal text']
    assert list(TextObject('normal text').wrap(5)) == ['norma', 'l tex', 't']
    assert list(TextObject('中文').wrap(1)) == ['中', '文']


def test_textobject_repeat_to():
    assert TextObject('-').repeat_to(5) == '-----'
    assert TextObject('XD').repeat_to(5) == 'XDXDX'
    assert TextObject('字').repeat_to(5) == '字字字'
    assert TextObject('.-*-').repeat_to(10) == '.-*-.-*-.-'



import pytest

from ..ascii_frame import wrap


def test_basic():
    assert wrap(['ascii_frame']) == [
        '.-----------.',
        '|ascii_frame|',
        "'-----------'",
    ]


def test_multi_lines():
    assert wrap(['ascii_frame', 'test']) == [
        '.-----------.',
        '|ascii_frame|',
        '|test       |',
        "'-----------'",
    ]


def test_width_narrow():
    assert wrap(['ascii_frame'], width=3) == [
        '.---.',
        '|asc|',
        '|ii_|',
        '|fra|',
        '|me |',
        "'---'",
    ]

    assert wrap(['ascii', 'frame'], width=3) == [
        '.---.',
        '|asc|',
        '|ii |',
        '|fra|',
        '|me |',
        "'---'",
    ]


def test_width_wide():
    assert wrap(['a', 'b'], width=5) == [
        '.-----.',
        '|a    |',
        '|b    |',
        "'-----'",
    ]


def test_padding_basic():
    assert wrap(['ascii', 'frame'], padding=3) == [
        '.-----------.',
        '|   ascii   |',
        '|   frame   |',
        "'-----------'",
    ]


def test_narrow_width_with_padding():
    assert wrap(['ascii_frame'], width=9, padding=3) == [
        '.---------.',
        '|   asc   |',
        '|   ii_   |',
        '|   fra   |',
        '|   me    |',
        "'---------'",
    ]

    assert wrap(['ascii', 'frame'], width=9, padding=3) == [
        '.---------.',
        '|   asc   |',
        '|   ii    |',
        '|   fra   |',
        '|   me    |',
        "'---------'",
    ]


def test_wide_width_with_padding():
    assert wrap(['a', 'b'], width=15, padding=3) == [
        '.---------------.',
        '|   a           |',
        '|   b           |',
        "'---------------'",
    ]


def test_corners_basic():
    assert wrap(['YATTA'], corners='+') == [
        '+-----+',
        '|YATTA|',
        '+-----+',
    ]

    assert wrap(['YATTA'], corners='+#') == [
        '+-----+',
        '|YATTA|',
        '#-----#',
    ]

    assert wrap(['YATTA'], corners='+#$%') == [
        '%-----+',
        '|YATTA|',
        '$-----#',
    ]


def test_edges_basic():
    assert wrap(['YATTA'], corners='#', edges='#') == [
        '#######',
        '#YATTA#',
        '#######',
    ]

    assert wrap(['YATTA'], corners='/\\/\\', edges='|-') == [
        '\|||||/',
        '-YATTA-',
        '/|||||\\',
    ]

    assert wrap(['YATTA'], corners='#', edges='*|*') == [
        '#*****#',
        '|YATTA|',
        '#*****#',
    ]

    assert wrap(['YATTA'], corners=' ', edges='^>v<') == [
        ' ^^^^^ ',
        '<YATTA>',
        ' vvvvv ',
    ]


def test_corners_and_edges():
    assert wrap(['YATTA'], padding=1, corners=['++'], edges=['=', '||'] * 2) == [
        '++=======++',
        '|| YATTA ||',
        '++=======++',
    ]


def test_fancy_edges():
    assert wrap(['YATTA YATTA'], padding=1, corners='.*', edges=['-*-.', '|', "-.-*"]) == [
        '.-*-.-*-.-*-.-.',
        '| YATTA YATTA |',
        '*-.-*-.-*-.-*-*',
    ]

    assert wrap(['YATTA YATTA YATTA!!!!'], padding=1,
            corners='  _`',
            edges=['"*-=+._`', ')', '.+=-*"`_', '>']) == [
        '`"*-=+._`"*-=+._`"*-=+._ ',
        '> YATTA YATTA YATTA!!!! )',
        '_.+=-*"`_.+=-*"`_.+=-*"` ',
    ]

    # wide edges
    assert wrap(['YATTA'], corners='十' * 4, edges='十' * 4) == [
        '十十十十十',
        '十YATTA 十',
        '十十十十十',
    ]


def test_exceptions_narrow():
    # too narrow
    with pytest.raises(ValueError):
        wrap(['a'], width=3, padding=3)


def test_exceptions_make_up_corners():
    # automatically make up corners
    wrap(['a'], corners='.')
    wrap(['a'], corners=['.'])
    wrap(['a'], corners='.' * 2)
    wrap(['a'], corners=['.'] * 2)

    # not enough corner
    with pytest.raises(ValueError):
        wrap(['a'], corners='.' * 3)
        wrap(['a'], corners=['.'] * 3)

    # right corners amount
    wrap(['a'], corners='.' * 4)
    wrap(['a'], corners=['.'] * 4)

    # too many corners
    with pytest.raises(ValueError):
        wrap(['a'], corners='.' * 5)
        wrap(['a'], corners=['.'] * 5)


def test_exceptions_make_up_edges():
    # automatically make up edges
    wrap(['a'], edges='.')
    wrap(['a'], edges=['.'])
    wrap(['a'], edges='.' * 2)
    wrap(['a'], edges=['.'] * 2)
    wrap(['a'], edges='.' * 3)
    wrap(['a'], edges=['.'] * 3)

    # right edges amount
    wrap(['a'], edges='.' * 4)
    wrap(['a'], edges=['.'] * 4)

    # too many edges
    with pytest.raises(ValueError):
        wrap(['a'], edges='.' * 5)
        wrap(['a'], edges=['.'] * 5)

    # wide corners
    with pytest.raises(ValueError):
        wrap(['a'], corners=['++'] * 4, edges=['-', '|'] * 2)

    # wide edges
    with pytest.raises(ValueError):
        wrap(['a'], corners=['+'] * 4, edges=['-', '||'] * 2)

def test_exceptions_wide_corner_and_edges():
    # wide corners with wide edges, ok
    wrap(['a'], corners=['++'] * 4, edges=['-', '||'] * 2)



import re
import math

from unicodedata import east_asian_width
# for more information, see http://unicode.org/reports/tr11/


class TextObject:
    @classmethod
    def uwidth(cls, c):
        return 1 + (east_asian_width(c) in 'WF')

    def __init__(self, text, rstrip=True):
        self.text = text.rstrip() if rstrip else text

    @property
    def width(self):
        return sum(
                1 + (east_asian_width(c) in 'WF')
                for c in re.sub(r'\x1b\[[^hm]*[hm]', '', self.text))

    def wrap(self, width):
        w, ret = 0, ''
        for c in self.text:
            if w + self.uwidth(c) > width and w > 0:
                yield TextObject(ret)
                w, ret = 0, ''

            w, ret = w + self.uwidth(c), ret + c

        yield TextObject(ret)

    def __str__(self):
        return self.text

    def __eq__(self, other):
        if isinstance(other, str):
            return self.text == other

        if isinstance(other, TextObject):
            return self.text == other.text

        return str(self) == str(other)

    def __mul__(self, multiplier):
        return self.text * multiplier

    def __add__(self, other):
        return TextObject(self.text + other)

    def repeat_to(self, goal_width):
        ret = TextObject(self.text * math.floor(goal_width / self.width))
        if ret.width == goal_width:
            return ret

        w = ret.width
        for c in self.text:
            if w >= goal_width:
                return ret
            w += self.uwidth(c)
            ret += c

        return ret



import sys
import ascii_frame

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name='ascii_frame',
    version=ascii_frame.__version__,
    description='Wrap text with an ASCII frame.',
    url='https://github.com/pi314/ascii_frame',
    author='Cychih',
    author_email='michael66230@gmail.com',
    maintainer='Cychih',
    maintainer_email='michael66230@gmail.com',
    license='WTFPL',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Chinese (Traditional)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Utilities',
    ],
    packages=find_packages(exclude=['scripts']),
    scripts=['scripts/ascii_frame'],

    tests_require=['pytest'],
    cmdclass={'test': PyTest},
)
